/*
 * Copyright (C) 2011 Michael Dippery <mdippery@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import <Foundation/Foundation.h>

@class MDPair;

/// Provides powerful array operations using blocks. The methods are inspired by and works similar to the corresponding ones in Smalltalk. More similar methods are available in the NSDictionary(RubyEnumerable) category. The Ruby Enumerable documentation explains all the methods with examples http://www.ruby-doc.org/core/classes/Enumerable.html 

@interface NSDictionary (SmalltalkCollections)

/// ---------------------------------------------------------------------
/// @name Manipulate arrays
/// ---------------------------------------------------------------------

/// Performs the block on all elements' value in the dictionary.
- (NSDictionary *)do:(void (^)(id obj))block;
/// Returns a new dictionary with the key-value pairs returned from invoking the block on the passed dictionary's values. In Ruby this method is also called map.
- (NSDictionary *)collect:(id (^)(id obj))block;

/// A dictionary version of collect/map that passes both key and value to the block.
- (NSDictionary *)collectValueAndKey:(id (^)(id key, id value))block;

/// ---------------------------------------------------------------------
/// @name Searching and filtering dictionaries
/// ---------------------------------------------------------------------

/// Returns a new dictionary with all the key-value pairs for which the values returns NO when evaluated by the block.
- (NSDictionary *)reject:(BOOL (^)(id obj))block;
/// Returns a new dictionary with all the key-value pairs for which the values returns YES when evaluated by the block.
- (NSDictionary *)select:(BOOL (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Extracting information from a dictionary
/// ---------------------------------------------------------------------

/**  Derrive a single object by iterating over the values of a dictionary. This could for example be used to sum all the elements of a collection. In Ruby this method is also called reduce.
 *  
 *  See an example for how to use this method in the [NSArray(SmalltalkCollections) inject:into:] documentation.
 *  @param initial A starting point object that will be passed as the memo object for the first block evaluation.
 *  @param block The block to be evaluated for each value in the dictionary. The value returned from the block will be the memo object for the next block evaluation. When the block is evaluated with the last object the returned value will be the method result.
 *  @result The value returned when evaluating the block on the value of the last element in the dictionary.
 */
- (id)inject:(id)initial into:(id (^)(id memo, id obj))block;

@end

/// Provides powerful dictionary operations using blocks. The methods are inspired by and works similar to the corresponding ones in Ruby. More similar methods are available in the NSDictionary(SmalltalkCollections) category. The Ruby Enumerable documentation explains all the methods with examples http://www.ruby-doc.org/core/classes/Enumerable.html 

@interface NSDictionary (RubyEnumerable)

/// ---------------------------------------------------------------------
/// @name Evaluating dictionaries
/// ---------------------------------------------------------------------

/// Evaluates the block on all elements' values and returns YES if the block returns YES for all values.
- (BOOL)all:(BOOL (^)(id obj))block;
/// Evaluates the block the elements' values and returns YES if the block returns YES for any value.
- (BOOL)any:(BOOL (^)(id obj))block;
/// Evaluates the block the elements' values and returns YES if the block returns YES for none of the values.
- (BOOL)none:(BOOL (^)(id obj))block;
/// Evaluates the block the elements' values and returns YES if the block returns YES for exactly one of the values.
- (BOOL)one:(BOOL (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Finding min and max values
/// ---------------------------------------------------------------------

/// Returns the the maximum valued element based on evaluating the NSComparator block.
- (id)max:(NSComparator)block;
/// Returns the the minimum valued element based on evaluating the NSComparator block.
- (id)min:(NSComparator)block;

/// ---------------------------------------------------------------------
/// @name Partititioning
/// ---------------------------------------------------------------------

/// Returns two dictionaries (wrapped in an MDPair), the first containing the key-value pairs for which the block evaluates to YES on the values, the second containing the rest.
- (MDPair *)partition:(BOOL (^)(id obj))block;

@end


