// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import "AppDelegate.h"
#import "ApplicationStateChart.h"
#import "SHPMessageRouter.h"
#import "ExpensesViewController.h"
#import "DebtsViewController.h"
#import "ExpenseDetailViewController.h"
#import "TabBarStateChart.h"
#import "SHPProxyOutputPort.h"
#import "SHPProxyInputPort.h"
#import "ExpenseDetailLib.h"
#import "AddExpenseViewController.h"
#import "AddExpenseLib.h"
#import "ModalLib.h"
#import "DebtsLib.h"
#import "SHPMessageWrapperComponent.h"
#import "SHPSocketServerStateChart.h"
#import "SHPBonjourServerStateChart.h"
#import "SHPBonjourBrowserStateChart.h"
#import "SHPMessageServerBridgeComponent.h"
#import "SHPMessageWrapperBridgeAdapter.h"
#import "LoginViewController.h"
#import "LoginLib.h"

@implementation AppDelegate {
    SHPEventPort<ApplicationStateChartEventProtocol> *_eventPort;
    SHPProxyInputPort<ModelComponentInputProtocol> *_expenseModelInputPort;
}

- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions {
    SHPMessageRouter *router = [SHPMessageRouter sharedInstance];

    // UI Statechart
    ApplicationStateChart *applicationStateChart = [[ApplicationStateChart alloc] init];

    // UI controllers
    UITabBarController *tabBarController = [UITabBarController new];
    UINavigationController *navigationController = [UINavigationController new];
    ExpensesViewController *expensesViewController = [ExpensesViewController new];
    ExpenseDetailViewController *expenseDetailViewController = [ExpenseDetailViewController new];
    DebtsViewController *debtsViewController = [DebtsViewController new];
    AddExpenseViewController *addExpenseViewController = [AddExpenseViewController new];
    LoginViewController *loginViewController = [LoginViewController new];

    // Setup UI controllers
    [navigationController setViewControllers:@[expensesViewController]];
    [tabBarController setViewControllers:@[navigationController, [[UINavigationController alloc] initWithRootViewController:debtsViewController]]];

    // UI Libraries
    TabBarLib *tabBarLib = [[TabBarLib alloc] initWithTabBarController:tabBarController];
    ExpensesNavigationBarLib *expensesNavigationBarLib = [[ExpensesNavigationBarLib alloc] initWithNavigationController:navigationController];
    ExpensesListLib *expensesListLib = [[ExpensesListLib alloc] initWithExpensesViewController:expensesViewController];
    ExpenseDetailLib *expenseDetailLib = [[ExpenseDetailLib alloc] initWithExpenseDetailViewController:expenseDetailViewController];
    AddExpenseLib *addExpenseLib = [[AddExpenseLib alloc] initWithAddExpenseViewController:addExpenseViewController];
    ModalLib *modalLib = [[ModalLib alloc] initWithTabBarController:tabBarController];
    DebtsLib *debtsLib = [[DebtsLib alloc] initWithDebtsViewController:debtsViewController];
    LoginLib *loginLib = [[LoginLib alloc] initWithLoginViewController:loginViewController];

    // Bind Statecharts to UI Libraries
    [applicationStateChart.tabBarStateChart setLib:tabBarLib];
    [applicationStateChart.expensesNavigationControllerStateChart setLib:expensesNavigationBarLib];
    [applicationStateChart.tabBarStateChart.modalStateChart setLib:modalLib];
    [applicationStateChart.tabBarStateChart.modalStateChart.addExpenseStateChart setLib:addExpenseLib];
    [applicationStateChart.expensesNavigationControllerStateChart.expensesListStateChart setLib:expensesListLib];
    [applicationStateChart.expensesNavigationControllerStateChart.expenseDetailStateChart setLib:expenseDetailLib];
    [applicationStateChart.debtsStateChart setLib:debtsLib];
    [applicationStateChart.tabBarStateChart.modalStateChart.loginStateChart setLib:loginLib];

    // Bind UI Libraries with controllers
    [expensesNavigationBarLib setViewControllers:@[expensesViewController, expenseDetailViewController]];
    [modalLib setViewControllers:@[[[UINavigationController alloc] initWithRootViewController:addExpenseViewController], loginViewController]];

    // Bind UI controllers to ui event port
    SHPEventPort<ApplicationStateChartEventProtocol> *uiEventPort = applicationStateChart.eventPort;
    [tabBarLib setEventPort:uiEventPort];
    [expensesNavigationBarLib setEventPort:uiEventPort];
    [expensesListLib setEventPort:uiEventPort];
    [expenseDetailLib setEventPort:uiEventPort];
    [addExpenseLib setEventPort:uiEventPort];
    [debtsLib setEventPort:uiEventPort];
    [loginLib setEventPort:uiEventPort];

    // Models
    ModelComponent *modelComponent = [ModelComponent new];

    // Channels
    [router addPublishSubscribeChannelFromOutputPort:modelComponent.outputPort toInputPort:applicationStateChart.dataInputPort];
    [router addPublishSubscribeChannelFromOutputPort:modelComponent.inputPort toInputPort:applicationStateChart.outputPort];

    // Start the engines!
    [applicationStateChart initialTransition];

    // Message bridge server
    SHPMessageWrapperBridgeAdapter *serverAdapter = [SHPMessageWrapperBridgeAdapter new];
    SHPMessageServerBridgeComponent *messageServerBridgeComponent = [[SHPMessageServerBridgeComponent alloc] init];
    SHPMessageWrapperComponent *channelWrapper = [[SHPMessageWrapperComponent alloc] initWithChannelName:@"stateChartInfoPort"];

    // Channels
    [router addPublishSubscribeChannelFromOutputPort:applicationStateChart.stateChartInfoPort toInputPort:channelWrapper.inputPort];
    [router addPublishSubscribeChannelFromOutputPort:channelWrapper.outputPort toInputPort:serverAdapter.inputPort];
    [router addPublishSubscribeChannelFromOutputPort:serverAdapter.outputPort toInputPort:messageServerBridgeComponent.messagesInputPort];

    // Start the engines!
    [messageServerBridgeComponent configure];
    [messageServerBridgeComponent.commandInputPort eventMessageServerBridgeStart];

    // Window
    self.window = [[UIWindow alloc] initWithFrame:[[UIScreen mainScreen] bounds]];
    self.window.backgroundColor = [UIColor whiteColor];
    self.window.rootViewController = tabBarController;
    [self.window makeKeyAndVisible];

    // Debug
    _eventPort = (SHPEventPort <ApplicationStateChartEventProtocol> *) applicationStateChart.eventPort;
    _expenseModelInputPort = modelComponent.inputPort;

    [[SHPKeyboard sharedInstance] setDelegate:self];
    [[SHPKeyboard sharedInstance] setupKeyboard];

    return YES;
}

- (void)keyboardKeyPressed:(NSDictionary *)keyDict {
    NSString *key = keyDict[@"key"];

    if([key isEqualToString:@"1"]) {
        [_eventPort expensesTabSelected];
    } else if([key isEqualToString:@"2"]) {
        [_eventPort debtsTabSelected];
    } else if ([key isEqualToString:@"r"]) {
        [_expenseModelInputPort fetchExpenses];
    }
}


@end