/*
 * Copyright (C) 2011 Michael Dippery <mdippery@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import <Foundation/Foundation.h>

@class MDPair;

/// Provides powerful array operations using blocks. The methods are inspired by and works similar to the corresponding ones in Smalltalk. More similar methods are available in the NSArray(RubyEnumerable) category. The Ruby Enumerable documentation explains all the methods with examples http://www.ruby-doc.org/core/classes/Enumerable.html 

@interface NSArray (SmalltalkCollections)
/// ---------------------------------------------------------------------
/// @name Manipulate arrays
/// ---------------------------------------------------------------------

/// Performs the block on all elements in the array.
- (NSArray *)do:(void (^)(id obj))block;

/// Returns a new array with the values returned from invoking the block on the passed array. In Ruby this method is also called map.
- (NSArray *)collect:(id (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Searching and filtering arrays
/// ---------------------------------------------------------------------

/// Returns the first element satisfying the block.
- (id)detect:(BOOL (^)(id obj))block;

/// Returns the first elements saisfying the block. If non the none block is invoked and the result of that is returned.
- (id)detect:(BOOL (^)(id obj))block ifNone:(id (^)(void))none;

/// Returns a new array with all the elements for which the block returns NO.
- (NSArray *)reject:(BOOL (^)(id obj))block;

/// Returns a new array with all the elements for which the block returns YES.
- (NSArray *)select:(BOOL (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Extracting information from an array
/// ---------------------------------------------------------------------

/**  Derrive a single object by iterating over an array. This could for example be used to sum all the elements of an array. In Ruby this method is also called reduce.
*  
*  Example showing how inject:into: can be used to sum an array of numbers
*	NSArray *numbers = [NSArray arrayWithObjects:[NSNumber numberWithInt:3], [NSNumber numberWithInt:5], [NSNumber numberWithInt:7], nil];
*	NSNumber *sum = [numbers inject:[NSNumber numberWithInt:0] into:(id)^(id memo, id obj) {
*		int intSum = [obj intValue] + [memo intValue];
*		return [NSNumber numberWithInt:intSum];
*	}];
*	NSLog(@"The sum of %@ is %d", numbers, [sum intValue]);
*  @param initial A starting point object that will be passed as the memo object for the first block evaluation.
*  @param block The block to be evaluated for each object in the array. The value returned from the block will be the memo object for the next block evaluation. When the block is evaluated with the last object the returned value will be the method result.
*  @result The value returned when evaluating the block on the last element in the array.
 */
- (id)inject:(id)initial into:(id (^)(id memo, id obj))block;

@end

/// Provides powerful array operations using blocks. The methods are inspired by and works similar to the corresponding ones in Ruby. More similar methods are available in the NSArray(SmalltalkCollections) category. The Ruby Enumerable documentation explains all the methods with examples http://www.ruby-doc.org/core/classes/Enumerable.html 

@interface NSArray (RubyEnumerable)

/// ---------------------------------------------------------------------
/// @name Evaluating arrays
/// ---------------------------------------------------------------------

/// Evaluates the block on all elements and returns YES if the block returns YES for all array elements.
- (BOOL)all:(BOOL (^)(id obj))block;
/// Evaluates the block on elements and returns YES if the block returns YES for any of the array elements.
- (BOOL)any:(BOOL (^)(id obj))block;
/// Evaluates the block on the elements and returns YES if the block returns NO for all array elements.
- (BOOL)none:(BOOL (^)(id obj))block;
/// Evaluates the block on the elements and returns YES if the block returns YES for exactly one element.
- (BOOL)one:(BOOL (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Discarding the first elements
/// ---------------------------------------------------------------------

/// Returns a new a new array where the first n elements are dropped.
- (NSArray *)drop:(NSUInteger)n;
/// Drops elements up to, but not including, the first element for which the block returns NO and returns an array containing the remaining elements.
- (NSArray *)dropWhile:(BOOL (^)(id obj))block;

/// ---------------------------------------------------------------------
/// @name Finding min and max elements
/// ---------------------------------------------------------------------

/// Returns the the maximum valued element based on evaluating the NSComparator block.
- (id)max:(NSComparator)block;
/// Returns the the minimum valued element based on evaluating the NSComparator block.
- (id)min:(NSComparator)block;

/// ---------------------------------------------------------------------
/// @name Partititioning
/// ---------------------------------------------------------------------

/// Returns two arrays (wrapped in an MDPair), the first containing the elements for which the block evaluates to YES, the second containing the rest.
- (MDPair *)partition:(BOOL (^)(id obj))block;
/// Returns the first n elements in a new array.
- (NSArray *)take:(NSUInteger)n;
/// Passes elements to the block until the block returns NO, then stops iterating and returns an array of all prior elements.
- (NSArray *)takeWhile:(BOOL (^)(id obj))block;

@end
