// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPOutputPort.h>
#import <SHPStateChart/SHPInputPort.h>
#import <SHPStateChart/SHPEventMessage.h>
#import <SHPStateChart/SHPStateChart.h>
#import <XCTest/XCTest.h>
#import <objc/runtime.h>
#import "SHPTestPort.h"
#import "SHPMessageRouter.h"
#import "SHPEventMessage.h"
#import "SHPChannel.h"
#import "SHPEventHandlingResult.h"
#import "SHPStateChartRepresentation.h"


enum {
    XCTAsyncTestCaseStatusUnknown = 0,
    XCTAsyncTestCaseStatusWaiting,
    XCTAsyncTestCaseStatusSucceeded,
    XCTAsyncTestCaseStatusFailed,
    XCTAsyncTestCaseStatusCancelled,
};
typedef NSUInteger XCTAsyncTestCaseStatus;

static void *kLoopUntil_Key = "LoopUntil_Key";
static void *kNotified_Key = "kNotified_Key";
static void *kNotifiedStatus_Key = "kNotifiedStatus_Key";
static void *kExpectedStatus_Key = "kExpectedStatus_Key";

@implementation SHPTestPort {
    NSArray *_recordings;
    NSString *_type;
}

- (void)setInputChannel:(SHPChannel *)inputChannel {
    [inputChannel receiverReady:self];
    _inputChannel = inputChannel;
}

- (void)channel:(SHPChannel *)channel processMessage:(SHPEventMessage *)message {

}

- (NSString *)name {
    return nil;
}

- (void)channel:(SHPChannel *)channel didDeliverMessage:(SHPEventMessage *)message {
    if(_inputChannel == channel) {
        dispatch_async(self.messageQueue, ^{
            NSInvocation *invocation = message.invocation;
            [invocation setTarget:self];
            [invocation invoke];
        });
    }
}

- (dispatch_queue_t)messageQueue {
    return dispatch_get_main_queue();
}

- (void)forwardInvocation:(NSInvocation *)invocation {
    [invocation retainArguments];
    SHPEventMessage *eventMessage = [[SHPEventMessage alloc] initWithEventType:invocation.selector invocation:invocation senderName:@"Test Port" parentMessage:nil];
    [self.channel postMessage:eventMessage];
}

- (NSArray *)stateRecordingsForEventBlock:(void (^)())eventBlock eventSelector:(SEL)selector {
    eventBlock();
    [self waitForStatus:XCTAsyncTestCaseStatusSucceeded timeout:10 type:NSStringFromSelector(selector)];

    return _recordings;
}

- (void)didAddInitialEventHandlingResult:(SHPEventHandlingResult *)eventHandlingResult {

}

- (void)didAddEventHandlingResult:(SHPEventHandlingResult *)eventHandlingResult {
    _recordings = eventHandlingResult.transitionRepresentations;

    [self notify:XCTAsyncTestCaseStatusSucceeded type:eventHandlingResult.event];
}

- (void)didUpdateStateChartRepresentation:(SHPStateChartRepresentation *)stateChartRepresentation {

}

#pragma mark - Testing

- (void)waitForStatus:(XCTAsyncTestCaseStatus)status timeout:(NSTimeInterval)timeout type:(NSString *)type {
    _type = type;
    self.notified = NO;
    self.expectedStatus = status;
    self.loopUntil = [NSDate dateWithTimeIntervalSinceNow:timeout];

    NSDate *dt = [NSDate dateWithTimeIntervalSinceNow:0.1];
    while (!self.notified && [self.loopUntil timeIntervalSinceNow] > 0) {
        [[NSRunLoop currentRunLoop] runMode:NSDefaultRunLoopMode
                                 beforeDate:dt];
        dt = [NSDate dateWithTimeIntervalSinceNow:0.1];
    }

    // Only assert when notified. Do not assert when timed out
    // Fail if not notified
    if (self.notified) {
        XCTAssertEqual(self.notifiedStatus, self.expectedStatus, @"Notified status does not match the expected status.");
    } else {
        XCTFail(@"Async test timed out.");
    }
}

- (void)waitForTimeout:(NSTimeInterval)timeout
{
    self.notified = NO;
    self.expectedStatus = XCTAsyncTestCaseStatusUnknown;
    self.loopUntil = [NSDate dateWithTimeIntervalSinceNow:timeout];

    NSDate *dt = [NSDate dateWithTimeIntervalSinceNow:0.1];
    while (!self.notified && [self.loopUntil timeIntervalSinceNow] > 0) {
        [[NSRunLoop currentRunLoop] runMode:NSDefaultRunLoopMode
                                 beforeDate:dt];
        dt = [NSDate dateWithTimeIntervalSinceNow:0.1];
    }
}

- (void)notify:(XCTAsyncTestCaseStatus)status type:(NSString *)type {
    if ([type isEqualToString:_type]) {
        self.notifiedStatus = status;
        // self.notified must be set at the last of this method
        self.notified = YES;
    }
}

#pragma nark - Object Association Helpers -

- (void) setAssociatedObject:(id)anObject key:(void*)key
{
    objc_setAssociatedObject(self, key, anObject, OBJC_ASSOCIATION_RETAIN_NONATOMIC);
}

- (id) getAssociatedObject:(void*)key
{
    id anObject = objc_getAssociatedObject(self, key);
    return anObject;
}

#pragma mark - Property Implementations -

- (NSDate*) loopUntil
{
    return [self getAssociatedObject:kLoopUntil_Key];
}

- (void) setLoopUntil:(NSDate*)value
{
    [self setAssociatedObject:value key:kLoopUntil_Key];
}

- (BOOL) notified
{
    NSNumber *valueNumber = [self getAssociatedObject:kNotified_Key];
    return [valueNumber boolValue];
}

- (void) setNotified:(BOOL)value
{
    NSNumber *valueNumber = [NSNumber numberWithBool:value];
    [self setAssociatedObject:valueNumber key:kNotified_Key];
}

- (XCTAsyncTestCaseStatus) notifiedStatus
{
    NSNumber *valueNumber = [self getAssociatedObject:kNotifiedStatus_Key];
    return [valueNumber integerValue];
}

- (void) setNotifiedStatus:(XCTAsyncTestCaseStatus)value
{
    NSNumber *valueNumber = [NSNumber numberWithUnsignedInteger:value];
    [self setAssociatedObject:valueNumber key:kNotifiedStatus_Key];
}

- (XCTAsyncTestCaseStatus) expectedStatus
{
    NSNumber *valueNumber = [self getAssociatedObject:kExpectedStatus_Key];
    return [valueNumber integerValue];
}

- (void) setExpectedStatus:(XCTAsyncTestCaseStatus)value
{
    NSNumber *valueNumber = [NSNumber numberWithUnsignedInteger:value];
    [self setAssociatedObject:valueNumber key:kExpectedStatus_Key];
}


@end