// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPOrState.h>
#import "SHPBonjourBrowserStateChart.h"
#import "SHPSocketConnectionInfo.h"
#import "SHPInputPort.h"

@protocol SHPBonjourBrowserStateChartEventProtocol <SHPBonjourBrowserCommandProtocol, SHPBonjourBrowserNotificationProtocol>
@optional
- (void)eventBonjourBrowserFoundService:(NSNetService *)service;
- (void)eventBonjourBrowserLostService:(NSNetService *)service;
- (void)eventBonjourBrowserDidResolveAddressForService:(NSNetService *)service;
@end

BASE_STATE(SHPBonjourBrowserStateChart)

#pragma mark - StateStopped
STATE(SHPBonjourBrowserStateChart, StateStopped) STATE_END

#pragma mark - StateSearch
STATE(SHPBonjourBrowserStateChart, StateSearch)

- (void)eventEntry:(SHPEventMessage *)eventMessage {

}

- (void)eventExit:(SHPEventMessage *)eventMessage {

}

- (void)eventBonjourBrowserFoundDomain:(NSString *)domain {
    [self.stateChart.dataOutputPort eventBonjourBrowserFoundDomain:domain];
}

- (void)eventBonjourBrowserFoundService:(NSNetService *)service {
    [service setDelegate:self.stateChart];
    [service resolveWithTimeout:10];
    [self.stateChart.services addObject:service];
}

- (void)eventBonjourBrowserLostDomain:(NSString *)domain {
    [self.stateChart.dataOutputPort eventBonjourBrowserLostDomain:domain];
}

- (void)eventBonjourBrowserLostService:(NSNetService *)service {
    SHPSocketConnectionInfo *connectionInfo = [SHPSocketConnectionInfo infoWithPort:(uint16_t) service.port host:service.hostName];
    [self.stateChart.dataOutputPort eventBonjourBrowserLostServiceWithConnectionInfo:connectionInfo];
}

- (void)eventBonjourBrowserDidResolveAddressForService:(NSNetService *)service {
    SHPSocketConnectionInfo *connectionInfo = [SHPSocketConnectionInfo infoWithPort:(uint16_t) service.port host:service.hostName];
    [self.stateChart.dataOutputPort eventBonjourBrowserFoundServiceWithConnectionInfo:connectionInfo];
    [service setDelegate:nil];
    [self.stateChart.services removeObject:service];
}

STATE_END

#pragma mark - StateStartingSearching
STATE(SHPBonjourBrowserStateChart, StateStartingSearching)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.browser searchForServicesOfType:self.stateChart.type inDomain:self.stateChart.domain];
}

- (void)eventExit:(SHPEventMessage *)eventMessage {

}

- (void)eventBonjourBrowserSearching {
    [self.stateChart.dataOutputPort eventBonjourBrowserSearching];
}

STATE_END

#pragma mark - StateSearching
STATE(SHPBonjourBrowserStateChart, StateSearching)STATE_END

#pragma mark - StateStopping
STATE(SHPBonjourBrowserStateChart, StateStopping)
- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.dataOutputPort eventBonjourBrowserStopping];
    [self.stateChart.browser stop];
}

- (void)eventExit:(SHPEventMessage *)eventMessage {

}

- (void)eventBonjourBrowserStopped:(SHPEventMessage *)eventMessage {
    [self.stateChart.dataOutputPort eventBonjourBrowserStopped];
}

STATE_END


@interface SHPBonjourBrowserStateChart () <NSNetServiceBrowserDelegate>
@property (nonatomic, readonly) SHPEventPort <SHPBonjourBrowserStateChartEventProtocol> *eventPort;
@end

@implementation SHPBonjourBrowserStateChart {

}

- (instancetype)initWithDomain:(NSString *)domain type:(NSString *)type {
    self = [super init];
    if (self) {

        _commandInputPort = [self inputPortWithProtocol:@protocol(SHPBonjourBrowserCommandProtocol) name:@"Command"];
        _dataOutputPort = [self outputPortWithProtocol:@protocol(SHPBonjourBrowserNotificationProtocol) name:@"Notification"];

        _domain = domain;
        _type = type;

        _services = [NSMutableArray new];

        _browser = [[NSNetServiceBrowser alloc] init];
        [_browser removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];
        [_browser scheduleInRunLoop:[NSRunLoop mainRunLoop] forMode:NSRunLoopCommonModes];
        _browser.delegate = self;
    }

    return self;
}

- (void)configureWithConfiguration:(id <SHPStateChartConfigurationProtocol>)configuration {
	StateStopped *stateStopped = [StateStopped new];
    StateSearch *stateSearch = [StateSearch new];
	StateStartingSearching *stateStartingSearching = [StateStartingSearching new];
	StateSearching *stateSearching = [StateSearching new];
	StateStopping *stateStopping = [StateStopping new];

    [stateSearch setSubStates:@[stateStartingSearching, stateSearching] initialState:stateStartingSearching];
    [configuration.rootState setSubStates:@[stateStopped, stateSearch, stateStopping] initialState:stateStopped];

    [stateStopped onEvent:@selector(eventBonjourBrowserSearch) transitionTo:stateSearch];

    [stateSearch onEvent:@selector(eventBonjourBrowserStop) transitionTo:stateStopping];
    [stateSearch internalTransitionOnEvent:@selector(eventBonjourBrowserFoundDomain:)];
    [stateSearch internalTransitionOnEvent:@selector(eventBonjourBrowserFoundService:)];
    [stateSearch internalTransitionOnEvent:@selector(eventBonjourBrowserLostDomain:)];
    [stateSearch internalTransitionOnEvent:@selector(eventBonjourBrowserLostService:)];
    [stateSearch internalTransitionOnEvent:@selector(eventBonjourBrowserDidResolveAddressForService:)];

    [stateStartingSearching onEvent:@selector(eventBonjourBrowserSearching) transitionTo:stateSearching];

    [stateStopping onEvent:@selector(eventBonjourBrowserStopped) transitionTo:stateStopped];


    [configuration setEventHandlerProtocol:@protocol(SHPBonjourBrowserStateChartEventProtocol)];
}

#pragma mark - Browser delegate

- (void)netServiceBrowser:(NSNetServiceBrowser *)aNetServiceBrowser didNotSearch:(NSDictionary *)errorDict {
    NSLog(@"Did not search");
}

- (void)netServiceBrowserWillSearch:(NSNetServiceBrowser *)aNetServiceBrowser {
    [self.eventPort eventBonjourBrowserSearching];
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)aNetServiceBrowser didFindDomain:(NSString *)domainString moreComing:(BOOL)moreComing {
    [self.eventPort eventBonjourBrowserFoundDomain:domainString];
}

- (void)netServiceBrowserDidStopSearch:(NSNetServiceBrowser *)aNetServiceBrowser {
    [self.eventPort eventBonjourBrowserStopped];
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)aNetServiceBrowser didFindService:(NSNetService *)aNetService moreComing:(BOOL)moreComing {
    [self.eventPort eventBonjourBrowserFoundService:aNetService];
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)aNetServiceBrowser didRemoveDomain:(NSString *)domainString moreComing:(BOOL)moreComing {
    [self.eventPort eventBonjourBrowserLostDomain:domainString];
}

- (void)netServiceBrowser:(NSNetServiceBrowser *)aNetServiceBrowser didRemoveService:(NSNetService *)aNetService moreComing:(BOOL)moreComing {
    [self.eventPort eventBonjourBrowserLostService:aNetService];
}


- (void)netServiceWillResolve:(NSNetService *)sender {
}

- (void)netServiceDidResolveAddress:(NSNetService *)netService {
    [self.eventPort eventBonjourBrowserDidResolveAddressForService:netService];
}

- (void)netService:(NSNetService *)sender didNotResolve:(NSDictionary *)errorDict {
    NSLog(@"Did not resolve");
}

@end
