// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPEventMessage.h>
#import "SHPBonjourServerStateChart.h"
#import "SHPInputPort.h"
#import "SHPOutputPort.h"
#import "NSError+ConvenienceCreatorAdditions.h"

@protocol SHPBonjourServerStateChartEventProtocol <SHPBonjourServerNotificationProtocol, SHPBonjourServerCommandProtocol>
@optional
@end

BASE_STATE(SHPBonjourServerStateChart)

#pragma mark - StateUnpublished
STATE(SHPBonjourServerStateChart, StateUnpublished)STATE_END

#pragma mark - StateWorking
STATE(SHPBonjourServerStateChart, StateWorking)

- (void)eventBonjourServerFailedWithError:(NSError *)error {
    [self.stateChart.notificationOutputPort eventBonjourServerFailedWithError:error];
}

STATE_END

#pragma mark - StatePublishing
STATE(SHPBonjourServerStateChart, StatePublishing)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.service publish];
}

- (void)eventBonjourServerPublishing {
    [self.stateChart.notificationOutputPort eventBonjourServerPublishing];
}

STATE_END

#pragma mark - StatePublished
STATE(SHPBonjourServerStateChart, StatePublished)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.notificationOutputPort eventBonjourServerPublished];
}

STATE_END

#pragma mark - StateStopping
STATE(SHPBonjourServerStateChart, StateStopping)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.notificationOutputPort eventBonjourServerStopping];
    [self.stateChart.service stop];
}

- (void)eventBonjourServerStopped {
    [self.stateChart.notificationOutputPort eventBonjourServerStopped];
}

STATE_END

#pragma mark - StateFailed
STATE(SHPBonjourServerStateChart, StateFailed)

STATE_END

@interface SHPBonjourServerStateChart () <NSNetServiceDelegate>
@property (nonatomic, readonly) SHPEventPort <SHPBonjourServerStateChartEventProtocol> *eventPort;
@end

@implementation SHPBonjourServerStateChart {

}

- (id)initWithDomain:(NSString *)domain type:(NSString *)type name:(NSString *)name port:(uint16_t)port {
    self = [super init];
    if (self) {
        _commandInputPort = [self inputPortWithProtocol:@protocol(SHPBonjourServerCommandProtocol) name:@"Command"];
        _notificationOutputPort = [self outputPortWithProtocol:@protocol(SHPBonjourServerNotificationProtocol) name:@"Notification"];

        _service = [[NSNetService alloc] initWithDomain:domain type:type name:name port:port];
        _service.delegate = self;
    }

    return self;
}


- (void)configureWithConfiguration:(id <SHPStateChartConfigurationProtocol>)configuration {
	StateUnpublished *stateUnpublished = [StateUnpublished new];
    StateWorking *stateWorking = [StateWorking new];
	StatePublishing *statePublishing = [StatePublishing new];
	StatePublished *statePublished = [StatePublished new];
    StateFailed *stateFailed = [StateFailed new];
    StateStopping *stateStopping = [StateStopping new];

    [stateWorking setSubStates:@[statePublishing, statePublished] initialState:statePublishing];
    [configuration.rootState setSubStates:@[stateUnpublished, stateWorking, stateFailed, stateStopping] initialState:stateUnpublished];


	[stateUnpublished onEvent:@selector(eventBonjourServerPublish) transitionTo:stateWorking];
    [statePublishing internalTransitionOnEvent:@selector(eventBonjourServerPublishing)];
	[statePublishing onEvent:@selector(eventBonjourServerPublished) transitionTo:statePublished];
    [stateWorking onEvent:@selector(eventBonjourServerFailedWithError:) transitionTo:stateFailed];
    [stateWorking onEvent:@selector(eventBonjourServerStop) transitionTo:stateStopping];
    [stateStopping onEvent:@selector(eventBonjourServerStopped) transitionTo:stateUnpublished];

    [configuration setEventHandlerProtocol:@protocol(SHPBonjourServerStateChartEventProtocol)];
}


#pragma mark - Service delegate

- (void)netServiceWillPublish:(NSNetService *)sender {
    [self.eventPort eventBonjourServerPublishing];
}

- (void)netServiceDidPublish:(NSNetService *)sender {
    [self.eventPort eventBonjourServerPublished];
}

- (void)netService:(NSNetService *)sender didNotPublish:(NSDictionary *)errorDict {
    NSError *error = nil;
    if(errorDict.count>0) {
        NSInteger errorCode = [errorDict.allKeys[0] intValue];
        NSString *errorDescription = errorDict.allValues[0];

        error = [NSError errorWithDescription:errorDescription code:errorCode];
    }

    [self.eventPort eventBonjourServerFailedWithError:error];
}

- (void)netServiceDidStop:(NSNetService *)sender {
    [self.eventPort eventBonjourServerStopped];
}

@end
