// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPLoggingUtilities.h>
#import "SHPMessageRouter.h"
#import "SHPUtilities.h"



@implementation SHPMessageRouter {
    dispatch_queue_t _queue;
    NSMutableDictionary *_channels;
    SHPChannel *_monitorChannel;
}

- (id)init {
    return [self initWithMonitorChannel:nil];
}

- (id)initWithMonitorChannel:(SHPChannel *)monitorChannel {
    self = [super initWithComponentRegisterChannel:nil];
    if (self) {
        _queue = SHPCreateSerialQueueWithRandomName();
        _channels = [NSMutableDictionary new];
        _monitorChannel = monitorChannel;
        _monitorPort = [self outputPortWithProtocol:@protocol(SHPMessageRouterMonitorPortProtocol) name:@"Monitor"];
    }

    return self;
}

- (void)setMonitorChannel:(SHPChannel *)monitorChannel {
    dispatch_async(_queue, ^{
        _monitorChannel = monitorChannel;
    });
}


+ (SHPMessageRouter *)sharedInstance {
    static dispatch_once_t once;
    static SHPMessageRouter *sharedFoo;
    dispatch_once(&once, ^ { sharedFoo = [[self alloc] init]; });
    return sharedFoo;
}

- (SHPChannel *)pointToPointChannel {
    SHPChannel *channel = [[SHPChannel alloc] initWithType:SHPChannelTypePointToPoint messageRouter:self];

    [self createMessageQueueForChannel:channel];

    return channel;
}

- (SHPChannel *)publishSubscribeChannel {
    SHPChannel *channel = [[SHPChannel alloc] initWithType:SHPChannelTypePublishSubscribe messageRouter:self];

    [self createMessageQueueForChannel:channel];

    return channel;
}

- (void)addPublishSubscribeChannelFromOutputPort:(SHPPort *)outputPort toInputPort:(SHPPort *)inputPort {
    outputPort.channel = inputPort.channel = [self publishSubscribeChannel];
}

- (void)addPointToPointChannelFromOutputPort:(SHPPort *)outputPort toInputPort:(SHPPort *)inputPort {
    outputPort.channel = inputPort.channel = [self pointToPointChannel];
}

- (void)createMessageQueueForChannel:(SHPChannel *)channel {
    dispatch_async(_queue, ^{
        NSMutableArray *messageQueue = [NSMutableArray new];
        [_channels setObject:messageQueue forKey:channel.identifier];
    });
}

- (void)sendMessage:(SHPEventMessage *)message channel:(SHPChannel *)channel {
    dispatch_async(_queue, ^{
        [_monitorChannel postMessage:message];

        switch (channel.type) {
            case SHPChannelTypePointToPoint: {
                [self postPointToPointMessage:message channel:channel];
                break;
            }
            case SHPChannelTypePublishSubscribe: {
                [self postPublishSubscribeMessage:message channel:channel];
                break;
            }
        }
    });
}

#pragma mark - Messaging (Private)

- (void)postPointToPointMessage:(id)message channel:(SHPChannel *)channel {
    NSMutableArray *messageQueue = [_channels objectForKey:channel.identifier];
    [messageQueue addObject:message];
    [self processNextPointToPointMessageForChannel:channel];
}

- (void)postPublishSubscribeMessage:(SHPEventMessage *)message channel:(SHPChannel *)channel {
    NSMutableArray *messageQueue = [_channels objectForKey:channel.identifier];
    [messageQueue addObject:message];
    [self processNextPublishSubscribeMessageForChannel:channel];
}

- (void)processNextPublishSubscribeMessageForChannel:(SHPChannel *)channel {
    dispatch_async(_queue, ^{

        NSMutableArray *messageQueue = [_channels objectForKey:channel.identifier];

        // ----------------
        // SHPStateChartGuard: No messages
        // ----------------
        if ([messageQueue count] == 0) {
            return;
        }

        // ----------------
        // Get message to process
        // ----------------
        SHPEventMessage *msg = [messageQueue objectAtIndex:0];
        [messageQueue removeObjectAtIndex:0];


        // ----------------
        // SHPStateChartGuard: No receivers
        // ----------------
        if ([channel.receivers count] == 0) {
            return;
        }

        // ----------------
        // Deliver msg to all receivers
        // ----------------
        for (id<SHPChannelDelegate> receiver in channel.receivers) {
            [receiver channel:channel didDeliverMessage:msg];
        }

        [self.monitorPort didProcessMessage:msg];

        // ----------------
        // Process next msg
        // ----------------
        [self processNextPublishSubscribeMessageForChannel:channel];
    });
}

- (void)processNextPointToPointMessageForChannel:(SHPChannel *)channel {
    dispatch_async(_queue, ^{

        NSMutableArray *messageQueue = [_channels objectForKey:channel.identifier];

        // ----------------
        // SHPStateChartGuard: No messages
        // ----------------
        if ([messageQueue count] == 0) {
            return;
        }

        // ----------------
        // SHPStateChartGuard: No receivers ready
        // ----------------
        if ([channel.receivers count] == 0) {
            return;
        }

        // ----------------
        // Get message to process
        // ----------------
        SHPEventMessage *msg = [messageQueue objectAtIndex:0];
        [messageQueue removeObjectAtIndex:0];

        // ----------------
        // Deliver msg to first ready receiver & remove receiver from ready list
        // ----------------
        id<SHPChannelDelegate> receiver = [channel.receivers objectAtIndex:0];
        [channel.receivers removeObjectAtIndex:0];

        [receiver channel:channel didDeliverMessage:msg];


        [self.monitorPort didProcessMessage:msg];

        // ----------------
        // Process next msg
        // ----------------
        [self processNextPointToPointMessageForChannel:channel];
    });
}

- (void)receiverReady:(id <SHPChannelDelegate>)receiver forChannel:(SHPChannel *)channel {
    dispatch_async(_queue, ^{
        switch (channel.type) {
            case SHPChannelTypePointToPoint:
                [channel.receivers addObject:receiver];
                [self processNextPointToPointMessageForChannel:channel];
                break;

            case SHPChannelTypePublishSubscribe:
                if ([channel.receivers containsObject:receiver]) {
                    return;
                }

                [channel.receivers addObject:receiver];
                break;

            default:
                NSAssert(NO, @"Invalid channel type");
                break;
        }
    });
}

@end