// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import "SHPSocketClientStateChart.h"
#import "GCDAsyncSocket.h"
#import "SHPSocketConnectionInfo.h"
#import "SHPStateChartGuard.h"
#import "SHPEventMessage.h"
#import "SHPInputPort.h"

@protocol SHPSocketClientStateChartEventProtocol <SHPSocketClientCommandProtocol, SHPSocketClientNotificationProtocol>
@optional
@end

@interface SHPSocketClientStateChart ()
@property (nonatomic, readonly) SHPEventPort <SHPSocketClientStateChartEventProtocol> *eventPort;
@end


BASE_STATE(SHPSocketClientStateChart)

#pragma mark - StateWaiting
STATE(SHPSocketClientStateChart, StateWaiting)

- (void)eventSocketClientConnectWithConnectionInfo:(SHPSocketConnectionInfo *)connectionInfo {
    SHPSocketConnection *socketConnection = [[SHPSocketConnection alloc] initWithHost:connectionInfo.host port:connectionInfo.port delegate:self.stateChart];;
    self.stateChart.socketConnection = socketConnection;
    [socketConnection connect];
}

STATE_END

#pragma mark - StateConnecting
STATE(SHPSocketClientStateChart, StateConnecting)

- (void)eventSocketClientFailedWithError:(NSError *)error {
    [self.stateChart.notificationOutputPort eventSocketClientFailedWithError:error];
}

STATE_END

#pragma mark - StateConnected
STATE(SHPSocketClientStateChart, StateConnected)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [super eventEntry:eventMessage];
    [self.stateChart.notificationOutputPort eventSocketClientConnected];
}

- (void)eventSocketClientSendObject:(id <NSCoding>)object {
    [self.stateChart.socketConnection sendObject:object];
}

- (void)eventSocketClientReceivedObject:(id <NSCoding>)object {
    [self.stateChart.notificationOutputPort eventSocketClientReceivedObject:object];
}

- (void)eventSocketClientDisconnected {
    [self.stateChart.notificationOutputPort eventSocketClientDisconnected];
    self.stateChart.socketConnection = nil;

    [self.stateChart.eventPort eventSocketClientDisconnected];
}

STATE_END

#pragma mark - StateDisconnecting
STATE(SHPSocketClientStateChart, StateDisconnecting)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.socketConnection disconnect];
}

- (void)eventSocketClientDisconnected {
    [self.stateChart.notificationOutputPort eventSocketClientDisconnected];
    self.stateChart.socketConnection = nil;

    [self.stateChart.eventPort eventSocketClientDisconnected];
}

STATE_END

#pragma mark - StateNotConnected
STATE(SHPSocketClientStateChart, StateNotConnected)STATE_END


@interface SHPSocketClientStateChart ()
@end

@implementation SHPSocketClientStateChart {

}

- (id)initWithSocket:(GCDAsyncSocket *)socket {
    self = [self init];

    _socketConnection = [[SHPSocketConnection alloc] initWithSocket:socket delegate:self];

    return self;
}

- (id)init {
    self = [super init];
    if (self) {
        _commandInputPort = [self inputPortWithProtocol:@protocol(SHPSocketClientCommandProtocol) name:@"Command"];
        _notificationOutputPort = [self outputPortWithProtocol:@protocol(SHPSocketClientNotificationProtocol) name:@"Notification"];
    }

    return self;
}

- (void)configureWithConfiguration:(id <SHPStateChartConfigurationProtocol>)configuration {
    StateNotConnected *stateNotConnected = [StateNotConnected new];
    StateWaiting *stateWaiting = [StateWaiting new];
	StateConnecting *stateConnecting = [StateConnecting new];
	StateConnected *stateConnected = [StateConnected new];
	StateDisconnecting *stateDisconnecting = [StateDisconnecting new];

    [stateNotConnected setSubStates:@[stateWaiting, stateConnecting, stateDisconnecting] initialState:stateWaiting];
    [configuration.rootState setSubStates:@[stateConnected, stateNotConnected] initialState:stateNotConnected];

	[stateWaiting onEvent:@selector(eventSocketClientConnectWithConnectionInfo:) transitionTo:stateConnecting guard:[SHPStateChartGuard guardWithCondition:^BOOL {
        return self.socketConnection == nil;
    }                                                                                                                                     guardDescription:@"If not connected"]];

    [stateWaiting onEvent:@selector(eventSocketClientConnect) transitionTo:stateConnected guard:[SHPStateChartGuard guardWithCondition:^BOOL {
        return self.socketConnection != nil;
    }                                                                                                                 guardDescription:@"If connected"]];

	[stateConnecting onEvent:@selector(eventSocketClientConnected) transitionTo:stateConnected];
    [stateConnecting onEvent:@selector(eventSocketClientFailedWithError:) transitionTo:stateWaiting];

    [stateDisconnecting onEvent:@selector(eventSocketClientDisconnected) transitionTo:stateWaiting];

    [stateConnected onEvent:@selector(eventSocketClientDisconnect) transitionTo:stateDisconnecting];
    [stateConnected onEvent:@selector(eventSocketClientDisconnected) transitionTo:stateNotConnected];
    [stateConnected internalTransitionOnEvent:@selector(eventSocketClientSendObject:)];
    [stateConnected internalTransitionOnEvent:@selector(eventSocketClientReceivedObject:)];

    [configuration setEventHandlerProtocol:@protocol(SHPSocketClientStateChartEventProtocol)];
}

- (void)socketConnection:(SHPSocketConnection *)socketConnection didConnectToHost:(NSString *)host port:(uint16_t)port {
    [self.eventPort eventSocketClientConnected];
}

- (void)socketConnection:(SHPSocketConnection *)socketConnection didDisconnectWithError:(NSError *)err {
    [self.eventPort eventSocketClientDisconnected];
}

- (void)socketConnection:(SHPSocketConnection *)socketConnection didFailWithError:(NSError *)err {
    [self.eventPort eventSocketClientFailedWithError:err];
}

- (void)socketConnection:(SHPSocketConnection *)socketConnection didReciveObject:(id)object {
    [self.eventPort eventSocketClientReceivedObject:object];
}

@end
