// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import "SHPSocketConnection.h"
#import "GCDAsyncSocket.h"
#import "NSError+ConvenienceCreatorAdditions.h"

#define SocketConnectionHeaderSize sizeof(SHPSocketConnectionHeader)

const NSTimeInterval SocketDefaultConnectionTimeout = -1.0;

const NSUInteger SocketConnectionHeaderTag = 1;
const NSUInteger SocketConnectionBodyTag = 2;
const NSUInteger SocketConnectionTypeMessage = 1;

SHPSocketConnectionHeader DataToHeader(NSData *data);

@interface SHPSocketConnection ()
@property(nonatomic, strong) NSString *host;
@property(nonatomic) NSTimeInterval timeout;
@property(nonatomic, strong) NSNetService *netService;
@property(nonatomic, strong) dispatch_queue_t queue;
@property(nonatomic, strong) GCDAsyncSocket *socket;
@end

@implementation SHPSocketConnection {
    SHPSocketConnectionHeader _lastHeader;
    uint16_t _port;
}

- (id)initWithHost:(NSString *)host port:(uint16_t)port delegate:(id<SHPSocketConnectionDelegate>)delegate {
    self = [super init];
    if (self) {
        _delegate = delegate;
        _port = port;
        self.host = host;
        self.queue = dispatch_queue_create("socket_queue", DISPATCH_QUEUE_SERIAL);
        self.timeout = SocketDefaultConnectionTimeout;
    }
    return self;
}

- (id)initWithSocket:(GCDAsyncSocket *)socket delegate:(id<SHPSocketConnectionDelegate>)delegate {
    self = [self init];
    if (self) {
        _delegate = delegate;
        self.socket = socket;
        self.socket.delegate = self;
        _port = self.socket.connectedPort;
        self.host = self.socket.connectedHost;
        self.timeout = SocketDefaultConnectionTimeout;

        [self.socket readDataToLength:SocketConnectionHeaderSize withTimeout:self.timeout tag:SocketConnectionHeaderTag];
    }
    return self;
}


- (void)dealloc
{
    if (self.netService.delegate == self) {
        self.netService.delegate = nil;
        [self.netService stop];
    }
    self.socket.delegate = nil;
}

#pragma mark - Connection

- (void)connect {
    NSError *error = nil;

    if (self.socket) {
        if ([self.delegate respondsToSelector:@selector(socketConnection:didFailWithError:)]) {
            error = [NSError errorWithDescription:@"Socket already connected"];
            [self.delegate socketConnection:self didFailWithError:error];
        }
    }

    if (self.netService && !self.host) {
        self.netService.delegate = self;
        [self.netService resolveWithTimeout:self.timeout];
    }

    // create the socket
    _socket = [[GCDAsyncSocket alloc] initWithDelegate:self delegateQueue:self.queue];
    [self.socket setIPv6Enabled:NO];

    // connect to host and port
    [self.socket connectToHost:self.host onPort:self.port withTimeout:self.timeout error:&error];
    if (error) {
        if ([self.delegate respondsToSelector:@selector(socketConnection:didFailWithError:)]) {
            [self.delegate socketConnection:self didFailWithError:error];
        }
        _socket = nil;
    }else {
        if ([self.delegate respondsToSelector:@selector(socketConnection:didConnectToHost:port:)]) {
            [self.delegate socketConnection:self didConnectToHost:self.host port:self.port];
        }
    }
}

- (void)disconnect;
{
    [self.socket disconnect];
    _socket = nil;
}

- (BOOL)connected;
{
    return (self.socket.connectedHost != nil);
}

#pragma mark - Send data

- (void)sendObject:(id<NSCoding>)object;
{
    SHPSocketConnectionHeader header;
    header.type = SocketConnectionTypeMessage;
    header.bodyLength = 0;

    [self sendHeader:header object:object];
}

- (void)sendHeader:(SHPSocketConnectionHeader)header object:(id<NSCoding>)object;
{
    NSAssert(self.socket, @"SocketConnection: attempted to send an object without being connected");

    // encode data
    NSData *bodyData = nil;
    if (object) {
        bodyData = [NSKeyedArchiver archivedDataWithRootObject:object];
        header.bodyLength = (UInt32)bodyData.length;
    }

    // send the header
    NSData *headerData = [NSData dataWithBytes:&header length:SocketConnectionHeaderSize];
    [self.socket writeData:headerData withTimeout:self.timeout tag:SocketConnectionHeaderTag];

    // send the body
    if (header.bodyLength > 0) [self.socket writeData:bodyData withTimeout:self.timeout tag:SocketConnectionBodyTag];
}

#pragma mark - NSNetServiceDelegate

// net service did resolve
- (void)netServiceDidResolveAddress:(NSNetService *)sender;
{
    _host = self.netService.hostName;
    _port = (uint16_t) self.netService.port;
    self.netService.delegate = nil;
    [self connect];
}

- (void)netService:(NSNetService *)sender didNotResolve:(NSDictionary *)errorDict {
    if ([self.delegate respondsToSelector:@selector(socketConnection:didFailWithError:)]) {
        NSError *error = [NSError errorWithDescription:@"Unable to resolve server"];
        [self.delegate socketConnection:self didFailWithError:error];
    }
}


#pragma mark - GCDAsycnSocketDelegate

- (void)socket:(GCDAsyncSocket *)sock didConnectToHost:(NSString *)host port:(uint16_t)port;
{
    // start reading length data
    [self.socket readDataToLength:SocketConnectionHeaderSize withTimeout:self.timeout tag:SocketConnectionHeaderTag];

    // inform delegate that we are connected
    if ([self.delegate respondsToSelector:@selector(socketConnection:didConnectToHost:port:)]) {
        [self.delegate socketConnection:self didConnectToHost:host port:port];
    }
}

- (void)socketDidDisconnect:(GCDAsyncSocket *)sock withError:(NSError *)error;
{
    if ([self.delegate respondsToSelector:@selector(socketConnection:didDisconnectWithError:)]) {
        [self.delegate socketConnection:self didDisconnectWithError:error];
    }
}

- (void)socket:(GCDAsyncSocket *)sock didReadData:(NSData *)data withTag:(long)tag;
{
    id object;
    switch(tag) {
        case SocketConnectionHeaderTag:
            _lastHeader = DataToHeader(data);

            [self.socket readDataToLength:_lastHeader.bodyLength withTimeout:self.timeout tag:SocketConnectionBodyTag];
            break;
        case SocketConnectionBodyTag:
            object = [NSKeyedUnarchiver unarchiveObjectWithData:data];

            if ([self.delegate respondsToSelector:@selector(socketConnection:didReciveObject:)]) {
                [self.delegate socketConnection:self didReciveObject:object];
            }

            [self.socket readDataToLength:SocketConnectionHeaderSize withTimeout:self.timeout tag:SocketConnectionHeaderTag];
            break;

        default:
            NSLog(@"SocketConnection: ignoring unknown tag: %ld", tag);
    }
}


SHPSocketConnectionHeader DataToHeader(NSData *data)
{
    assert(data.length == 3 * sizeof(UInt32));

    const UInt32 *encodedHeader = (const UInt32 *)data.bytes;
    SHPSocketConnectionHeader header;
    header.type       = CFSwapInt16LittleToHost((uint16_t) encodedHeader[0]);
    header.blockTag   = CFSwapInt32LittleToHost(encodedHeader[1]);
    header.bodyLength = CFSwapInt32LittleToHost(encodedHeader[2]);
    return header;
}

@end