// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPMessageRouter.h>
#import "SHPSocketServerStateChart.h"
#import "SHPEventMessage.h"
#import "GCDAsyncSocket.h"
#import "SHPSocketClientStateChart.h"



@protocol SHPSocketServerStateChartEventProtocol <SHPSocketServerNotificationProtocol, SHPSocketServerCommandProtocol, SHPSocketClientNotificationProtocol>
@optional
- (void)eventSocketServerClientDisconnectedWithSocket:(GCDAsyncSocket *)socket error:(NSError *)error;
- (void)eventSocketServerClientConnectedWithSocket:(GCDAsyncSocket *)socket;
@end

@interface SHPSocketServerStateChart ()
@property (nonatomic, readonly) SHPEventPort <SHPSocketServerStateChartEventProtocol> *eventPort;
@end

BASE_STATE(SHPSocketServerStateChart)

#pragma mark - StateNotStarted
STATE(SHPSocketServerStateChart, StateNotStarted)STATE_END

#pragma mark - StateWaiting
STATE(SHPSocketServerStateChart, StateWaiting) STATE_END

#pragma mark - StateStarting
STATE(SHPSocketServerStateChart, StateStarting)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    NSError *error = nil;
    [self.stateChart.serverSocket acceptOnInterface:self.stateChart.interface port:self.stateChart.port error:&error];
    if(error) {
        [self.stateChart.eventPort eventSocketServerFailedWithError:error];
    }else {
        [self.stateChart.eventPort eventSocketServerStarted];
    }
}

- (void)eventSocketServerFailedWithError:(NSError *)error {
    [self.stateChart.notificationOutputPort eventSocketServerFailedWithError:error];
}

STATE_END

#pragma mark - StateStopping
STATE(SHPSocketServerStateChart, StateStopping)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.serverSocket disconnect];
}

- (void)eventSocketServerStopped {
    [self.stateChart.notificationOutputPort eventSocketServerStopped];
}

STATE_END

#pragma mark - StateFailed
STATE(SHPSocketServerStateChart, StateFailed)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [super eventEntry:eventMessage];
}


STATE_END

#pragma mark - StateStarted
STATE(SHPSocketServerStateChart, StateStarted)

- (void)eventEntry:(SHPEventMessage *)eventMessage {
    [self.stateChart.notificationOutputPort eventSocketServerStarted];
}

- (void)eventSocketClientReceivedObject:(id <NSCoding>)object {
    [self.stateChart.notificationOutputPort eventSocketServerReceivedObject:object];
}

- (void)eventSocketServerSendObject:(id <NSCoding>)object {
    [self.stateChart.commandOutputPort eventSocketClientSendObject:object];
}


- (void)eventSocketServerClientDisconnectedWithSocket:(GCDAsyncSocket *)socket error:(NSError *)error {

}

- (void)eventSocketServerClientConnectedWithSocket:(GCDAsyncSocket *)socket {
    SHPSocketClientStateChart *socketClientStateChart = [[SHPSocketClientStateChart alloc] initWithSocket:socket];
    socketClientStateChart.commandInputPort.channel = self.stateChart.commandOutputPort.channel;
    socketClientStateChart.notificationOutputPort.channel = self.stateChart.notificationInputPort.channel;
    [socketClientStateChart initialTransition];

    [self.stateChart.commandOutputPort eventSocketClientConnect];

    [self.stateChart addSubComponent:socketClientStateChart];
}

STATE_END


@implementation SHPSocketServerStateChart {
}

- (instancetype)initWithPort:(uint16_t)port interface:(NSString *)interface {
    self = [super init];
    if (self) {
        _port = port;
        _interface = interface;

        _commandInputPort = [self inputPortWithProtocol:@protocol(SHPSocketServerCommandProtocol) name:@"Command"];
        _notificationOutputPort = [self outputPortWithProtocol:@protocol(SHPSocketServerNotificationProtocol) name:@"Notification"];

        _commandOutputPort = [self outputPortWithProtocol:@protocol(SHPSocketClientCommandProtocol) name:@"Internal command"];
        _notificationInputPort = [self inputPortWithProtocol:@protocol(SHPSocketClientNotificationProtocol) name:@"Internal notification"];

        _commandOutputPort.channel = [[SHPMessageRouter sharedInstance] publishSubscribeChannel];
        _notificationInputPort.channel = [[SHPMessageRouter sharedInstance] publishSubscribeChannel];
    }

    return self;
}

- (void)configureWithConfiguration:(id <SHPStateChartConfigurationProtocol>)configuration {
    StateNotStarted *stateNotStarted = [StateNotStarted new];
    StateWaiting *stateWaiting = [StateWaiting new];
    StateStarting *stateStarting = [StateStarting new];
    StateStopping *stateStopping = [StateStopping new];
    StateFailed *stateFailed = [StateFailed new];
    StateStarted *stateStarted = [StateStarted new];

    [stateNotStarted setSubStates:@[stateWaiting, stateStopping, stateFailed, stateStarting] initialState:stateWaiting];
    [configuration.rootState setSubStates:@[stateNotStarted, stateStarted] initialState:stateNotStarted];

    [stateWaiting onEvent:@selector(eventSocketServerStart) transitionTo:stateStarting];
    [stateStopping onEvent:@selector(eventSocketServerStopped) transitionTo:stateWaiting];

    [stateStarting onEvent:@selector(eventSocketServerStarted) transitionTo:stateStarted];
    [stateStarting onEvent:@selector(eventSocketServerFailedWithError:) transitionTo:stateFailed];

    [stateStarted internalTransitionOnEvent:@selector(eventSocketServerSendObject:)];
    [stateStarted internalTransitionOnEvent:@selector(eventSocketClientReceivedObject:)];
    [stateStarted internalTransitionOnEvent:@selector(eventSocketServerClientConnectedWithSocket:)];
    [stateStarted internalTransitionOnEvent:@selector(eventSocketServerClientDisconnectedWithSocket:error:)];

    [stateStarted onEvent:@selector(eventSocketServerStop) transitionTo:stateStopping];

    [configuration setEventHandlerProtocol:@protocol(SHPSocketServerStateChartEventProtocol)];

    _serverSocket = [[GCDAsyncSocket alloc] initWithDelegate:self delegateQueue:self.messageQueue];
    self.serverSocket.delegate = self;
}

#pragma mark - AsyncSocket delegate

- (void)socket:(GCDAsyncSocket *)socket didAcceptNewSocket:(GCDAsyncSocket *)newSocket {
    [self.eventPort eventSocketServerClientConnectedWithSocket:newSocket];
}

- (void)socketDidDisconnect:(GCDAsyncSocket *)socket withError:(NSError *)error {
    [self.eventPort eventSocketServerClientDisconnectedWithSocket:socket error:error];
}

@end
