// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.


#import "SHPEventRepresentation.h"
#import <SHPStateChart/SHPStateChart.h>
#import <SHPStateChart/SHPEventRepresentation.h>
#import "SHPAndState.h"
#import "SHPEventMessage.h"
#import "NSArrayCollections.h"
#import "SHPStateChartTransition.h"
#import "SHPOrState.h"
#import "SHPEventHandlingResult.h"
#import "SHPAndStateRepresentation.h"
#import "SHPStateRepresentation.h"
#import "SHPTransitionRepresentation.h"

@interface SHPAndState ()

@end

@implementation SHPAndState

#pragma mark - Statechart

- (void)setStateChartUnlessSetRecursively:(SHPStateChart *)statechart {
    if (!_stateChart) {
        _stateChart = statechart;
        for (SHPState *state in _regions) {
            [state setStateChartUnlessSetRecursively:statechart];
        }
    }
}

#pragma mark - State

- (SHPState *)findFirstStateOfClassRecursively:(Class)stateClass {
    if ([self isKindOfClass:stateClass]) {
        return self;
    }

    for (SHPState *state in _regions) {
        SHPState *foundState = [state findFirstStateOfClassRecursively:stateClass];
        if (foundState) { return foundState; }
    }

    return nil;
}

#pragma mark - Transitions

- (void)exitCurrentConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)eventTransitionRepresentation {
    for (SHPOrState *region in _regions) {
        [region exitCurrentConfigurationWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
    }

    SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeExit) transitionAction:nil stateRepresentation:[self representationForState]];
    [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

    [self removeTimeoutTimers];
    [self eventExit:eventMessage];
}

- (void)enterTargetConfigurationWithTargetStates:(NSArray *)targetStates eventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf historyMode:(HistoryMode)historyMode transitionRepresentation:(SHPTransitionRepresentation *)eventTransitionRepresentation {
    if (enterSelf) {
        SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
        [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

        [self setupTimeoutTimers];
        [self eventEntry:eventMessage];
    }

    for (SHPOrState *region in _regions) {
        if ([region isOrContainsTargetStates:targetStates]) {
            [region enterTargetConfigurationWithTargetStates:targetStates eventMessage:eventMessage enterSelf:YES historyMode:historyMode transitionRepresentation:eventTransitionRepresentation];
        } else {
            switch (historyMode) {
                case HistoryModeNone:
                case HistoryModeShallow:
                    [region enterDefaultConfigurationWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
                    break;
                case HistoryModeDeep:
                    [region enterDeepHistoryWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
                    break;
            }
        }
    }
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    [self enterDefaultConfigurationWithEventMessage:eventMessage enterSelf:YES transitionRepresentation:transitionRepresentation];
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    if (enterSelf) {
        SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
        [transitionRepresentation addEventRepresentation:eventRepresentation];

        [self setupTimeoutTimers];
        [self eventEntry:eventMessage];
    }

    for (SHPOrState *region in _regions) {
        [region enterDefaultConfigurationWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
    }
}

- (void)enterDeepHistoryWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
    [transitionRepresentation addEventRepresentation:eventRepresentation];

    [self setupTimeoutTimers];
    [self eventEntry:eventMessage];

    for (SHPOrState *region in _regions) {
        [region enterDeepHistoryWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
    }
}

- (BOOL)isOrContainsTargetStates:(NSArray *)targetStates {
    if ([[targetStates select:^BOOL(SHPState *state) { return self == state; }] count] > 0) { return YES; }

    for (SHPOrState *region in _regions) {
        if ([region isOrContainsTargetStates:targetStates]) {
            return YES;
        }
    }

    return NO;
}

- (NSOrderedSet *)transitionsRespondingToEventMessage:(SHPEventMessage *)eventMessage currentPerformingTransition:(SHPStateChartTransition *)currentPerformingTransition {
    SHPStateChartTransition *transition = [[self transitionsForEvent:eventMessage.eventType checkCondition:YES] firstObject];
    currentPerformingTransition = transition ? transition : currentPerformingTransition;

    NSMutableOrderedSet *respondingTransitions = [NSMutableOrderedSet orderedSetWithCapacity:[self.regions count]];
    for (SHPOrState *regionState in self.regions) {
        NSOrderedSet *regionRespondingTransitions = [regionState transitionsRespondingToEventMessage:eventMessage currentPerformingTransition:nil];

        if (regionRespondingTransitions) {[respondingTransitions unionOrderedSet:regionRespondingTransitions]; }
    }

    NSOrderedSet *flattenedRespondingTransitions = [respondingTransitions copy];

    if ([flattenedRespondingTransitions count] > 0) {
        return flattenedRespondingTransitions;
    } else if (currentPerformingTransition != nil) {
        return [NSOrderedSet orderedSetWithObject:currentPerformingTransition];
    } else {
        return [NSOrderedSet orderedSet];
    }
}

- (void)setRegions:(NSArray *)regions {
    _regions = regions;
    for (SHPOrState *region in _regions) {
        region.superstate = self;
    }
}

- (SHPAndStateRepresentation *)representationForState {
    SHPAndStateRepresentation *representationAndState = [[SHPAndStateRepresentation alloc] init];
    representationAndState.shortName = [self name];
    representationAndState.uniqueName = NSStringFromClass([self class]);
    representationAndState.active = YES;

    representationAndState.regions = [self.regions collect:^id(SHPState *region) {
        SHPStateRepresentation *representationState = [region representationForState];
        return representationState;
    }];

    return representationAndState;
}

@end
