// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import <SHPStateChart/SHPEventRepresentation.h>
#import "SHPEventRepresentation.h"
#import "SHPOrState.h"
#import "SHPStateChartTransition.h"
#import "SHPStateChart.h"
#import "NSArrayCollections.h"
#import "SHPEventMessage.h"
#import "SHPEventHandlingResult.h"
#import "SHPOrStateRepresentation.h"
#import "SHPTransitionRepresentation.h"

@implementation SHPOrState

#pragma mark - Statechart

- (void)setStateChartUnlessSetRecursively:(SHPStateChart *)statechart {
    if (!_stateChart) {
        _stateChart = statechart;
        for (SHPState *state in _subStates) {
            [state setStateChartUnlessSetRecursively:statechart];
        }
    }
}

#pragma mark - State

- (SHPState *)findFirstStateOfClassRecursively:(Class)stateClass {
    if ([self isKindOfClass:stateClass]) {
        return self;
    }

    for (SHPState *state in _subStates) {
        SHPState *foundState = [state findFirstStateOfClassRecursively:stateClass];
        if (foundState) { return foundState; }
    }

    return nil;
}

#pragma mark - Setup

- (void)setSubStates:(NSArray *)subStates initialState:(SHPState *)initialState {
    _subStates = subStates;
    _initialState = initialState;
    _historyState = initialState;

    for (SHPState *state in subStates) {
        state.superstate = self;
    };
}

#pragma mark - Inspection

- (NSOrderedSet *)transitionsRespondingToEventMessage:(SHPEventMessage *)eventMessage currentPerformingTransition:(SHPStateChartTransition *)currentPerformingTransition {
    SHPStateChartTransition *transition = [[self transitionsForEvent:eventMessage.eventType checkCondition:YES] firstObject];
    currentPerformingTransition = transition ? transition : currentPerformingTransition;

    if ([_subStates count] == 0 && currentPerformingTransition != nil) {
        return [NSOrderedSet orderedSetWithArray:@[ currentPerformingTransition ]];
    } else if ([_subStates count] == 0) {
        return [NSOrderedSet orderedSetWithArray:@[ ]];
    } else {
        return [_activeState transitionsRespondingToEventMessage:eventMessage currentPerformingTransition:currentPerformingTransition];
    }
}

- (void)exitCurrentConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)eventTransitionRepresentation {
    if ([_subStates count] > 0) {
        [_activeState exitCurrentConfigurationWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
        _historyState = _activeState;
        _activeState = nil;
    }

    SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeExit) transitionAction:nil stateRepresentation:[self representationForState]];
    [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

    [self removeTimeoutTimers];
    [self eventExit:eventMessage];
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    [self enterDefaultConfigurationWithEventMessage:eventMessage enterSelf:YES transitionRepresentation:transitionRepresentation];
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    if (enterSelf) {
        SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
        [transitionRepresentation addEventRepresentation:eventRepresentation];

        [self setupTimeoutTimers];
        [self eventEntry:eventMessage];
    }

    if ([self isCompositeState]) {
        SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeInitial) transitionAction:nil stateRepresentation:[self representationForState]];
        [transitionRepresentation addEventRepresentation:eventRepresentation];

        [self eventInitial:eventMessage];
        [[self initialState] enterDefaultConfigurationWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
        _activeState = _initialState;
    }
}

- (void)enterDeepHistoryWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
    [transitionRepresentation addEventRepresentation:eventRepresentation];

    [self setupTimeoutTimers];
    [self eventEntry:eventMessage];

    [_historyState enterDeepHistoryWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
    _activeState = _historyState;
}

- (BOOL)isCompositeState {
    return [_subStates count] > 0;
}

- (void)enterTargetConfigurationWithTargetStates:(NSArray *)targetStates eventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf historyMode:(HistoryMode)historyMode transitionRepresentation:(SHPTransitionRepresentation *)eventTransitionRepresentation {
    if (enterSelf) {
        SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeEntry) transitionAction:nil stateRepresentation:[self representationForState]];
        [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

        [self setupTimeoutTimers];
        [self eventEntry:eventMessage];
    }

    for (SHPOrState *subState in _subStates) {
        if ([subState isOrContainsTargetStates:targetStates]) {
            _activeState = subState;
            [subState enterTargetConfigurationWithTargetStates:targetStates eventMessage:eventMessage enterSelf:YES historyMode:historyMode transitionRepresentation:eventTransitionRepresentation];
        }
    }

    if ([[targetStates select:^BOOL(SHPOrState *state) { return self == state; }] count] > 0) {
        switch (historyMode) {
            case HistoryModeNone: {
                SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeInitial) transitionAction:nil stateRepresentation:[self representationForState]];
                [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

                [self eventInitial:eventMessage];
                [[self initialState] enterDefaultConfigurationWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
                _activeState = _initialState;
                break;
            }
            case HistoryModeShallow: {
                SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeInitial) transitionAction:nil stateRepresentation:[self representationForState]];
                [eventTransitionRepresentation addEventRepresentation:eventRepresentation];

                [self eventInitial:eventMessage];
                [_historyState enterDefaultConfigurationWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
                _activeState = _historyState;
                break;
            }
            case HistoryModeDeep:
                [_historyState enterDeepHistoryWithEventMessage:eventMessage transitionRepresentation:eventTransitionRepresentation];
                _activeState = _historyState;
                break;
        }
    }
}

- (BOOL)isOrContainsTargetStates:(NSArray *)states {
    if ([[states select:^BOOL(SHPOrState *state) { return self == state; }] count] > 0) { return YES; }

    for (SHPOrState *subState in _subStates) {
        if ([subState isOrContainsTargetStates:states]) {
            return YES;
        }
    }

    return NO;
}

- (SHPOrStateRepresentation *)representationForState {
    SHPOrStateRepresentation *representationOrState = [[SHPOrStateRepresentation alloc] init];
    representationOrState.shortName = [self name];
    representationOrState.uniqueName = NSStringFromClass([self class]);
    representationOrState.subStates = [self.subStates collect:^id(SHPState *subState) {
        SHPStateRepresentation *representationState = [subState representationForState];
        if (subState == self.activeState) {
            representationState.active = YES;
        }
        return representationState;
    }];

    return representationOrState;
}

@end
