// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.

#import "SHPEventRepresentation.h"
#import <SHPStateChart/SHPStateChart.h>
#import "SHPState.h"
#import "SHPOrState.h"
#import "SHPEventMessage.h"
#import "SHPStateChartTransition.h"
#import "SHPStateChartGuardTransition.h"
#import "SHPStateChartGuard.h"
#import "SHPEventHandlingResult.h"
#import "SHPStateChartTimeoutTransition.h"
#import "SHPStateChartDeferredTransition.h"
#import "MSWeakTimer.h"
#import "SHPOrStateRepresentation.h"
#import "SHPStateRepresentation.h"
#import "SHPTransitionRepresentation.h"
#import "NSArrayCollections.h"
#import "NSArray+GroupingAdditions.h"

@interface SHPState ()

- (void)addEventTransition:(SHPStateChartTransition *)transition;

@end

@implementation SHPState {
    NSMutableDictionary *_deferEvents;
    NSArray *_timeoutTimers;
}

#pragma mark - Init

- (id)init {
    self = [super init];
    if (self) {
        _events = [NSMutableDictionary dictionary];
        _deferEvents = [NSMutableDictionary dictionary];
    }

    return self;
}

#pragma mark - Properties

- (NSString *)name {
    return NSStringFromClass([self class]);
}

#pragma mark - Statechart

- (void)setStateChartUnlessSetRecursively:(SHPStateChart *)statechart {
    NSAssert(false, @"Must implement in subclass");
}

#pragma mark - State
- (SHPState *)findFirstStateOfClassRecursively:(Class)stateClass {
    NSAssert(false, @"Must implement in subclass");

    return nil;
}

#pragma mark - Transitions

+ (void)onEvent:(SEL)event fromRegionStates:(NSArray *)fromRegionStates transitionTo:(NSArray *)toRegionStates {
    SHPStateChartTransition *transition = [SHPStateChartTransition transitionWithEvent:event fromStates:fromRegionStates toStates:toRegionStates timeout:nil internal:NO guard:nil historyMode:(HistoryModeShallow)];
    for (SHPOrState *state in fromRegionStates) {
        [state addEventTransition:transition];
    }
}

- (void)onEvent:(SEL)event transitionToShallowHistoryOf:(SHPState *)toState {
    SHPStateChartTransition *transition = [SHPStateChartTransition transitionWithEvent:event fromStates:@[ self ] toStates:@[ toState ] timeout:nil internal:NO guard:nil historyMode:HistoryModeShallow];
    [self addEventTransition:transition];
}

- (void)onEvent:(SEL)event transitionToDeepHistoryOf:(SHPState *)toState {
    SHPStateChartTransition *transition = [SHPStateChartTransition transitionWithEvent:event fromStates:@[ self ] toStates:@[ toState ] timeout:nil internal:NO guard:nil historyMode:HistoryModeDeep];
    [self addEventTransition:transition];
}

- (void)transitionTo:(SHPState *)toState {
    [self transitionTo:toState action:NULL];
}

- (void)transitionTo:(SHPState *)toState action:(SEL)action {
    SHPStateChartGuardTransition *transition = [SHPStateChartGuardTransition transitionWithEvent:@selector(guardTransition) action:action fromStates:@[self] toStates:@[toState] timeout:nil internal:NO guard:nil];
    [self addEventTransition:transition];
}

- (void)transitionTo:(SHPState *)toState guard:(SHPStateChartGuard *)guard {
    SHPStateChartGuardTransition *transition = [SHPStateChartGuardTransition transitionWithEvent:@selector(guardTransition) action:NULL fromStates:@[self] toStates:@[toState] timeout:nil internal:NO guard:guard];
    [self addEventTransition:transition];
}

- (void)transitionTo:(SHPState *)toState guard:(SHPStateChartGuard *)guard reverse:(BOOL)reverse {
    [self transitionTo:toState action:NULL guard:guard reverse:reverse];
}

- (void)transitionTo:(SHPState *)toState action:(SEL)action guard:(SHPStateChartGuard *)guard {
    [self transitionTo:toState action:action guard:guard reverse:NO];
}

- (void)transitionTo:(SHPState *)toState action:(SEL)action guard:(SHPStateChartGuard *)guard reverse:(BOOL)reverse {
    SHPStateChartGuardTransition *transition = [SHPStateChartGuardTransition transitionWithEvent:@selector(guardTransition) action:action fromStates:@[self] toStates:@[toState] timeout:nil internal:NO guard:guard];
    [self addEventTransition:transition];

    if(reverse) {
        SHPStateChartGuard *reversedGuard = [[SHPStateChartGuard alloc] initWithCondition:^BOOL{
            return !guard.condition();
        } guardDescription:[NSString stringWithFormat:@"Reverse of %@", guard.guardDescription]];
        SHPStateChartGuardTransition *backTransition = [SHPStateChartGuardTransition transitionWithEvent:@selector(guardTransition) action:action fromStates:@[toState] toStates:@[self] timeout:nil internal:NO guard:reversedGuard];
        [toState addEventTransition:backTransition];
    }
}

- (void)onEvent:(SEL)event transitionTo:(SHPState *)toState {
    [self onEvent:event transitionTo:toState guard:nil];
}

- (void)onEvent:(SEL)event transitionTo:(SHPState *)toState guard:(SHPStateChartGuard *)guard {
    [self onEvent:event transitionTo:toState guard:guard timeout:nil];
}

- (void)onEvent:(SEL)event transitionTo:(SHPState *)toState guard:(SHPStateChartGuard *)guard timeout:(NSNumber *)timeout {
    [self onEvent:event transitionToStates:@[ toState ] guard:guard];
}

- (void)onEvent:(SEL)event transitionToStates:(NSArray *)toStates guard:(SHPStateChartGuard *)guard {
    SHPStateChartTransition *transition = [SHPStateChartTransition transitionWithEvent:event fromStates:@[ self ] toStates:toStates timeout:nil internal:NO guard:guard historyMode:HistoryModeNone];
    [self addEventTransition:transition];
}

- (void)addEventTransition:(SHPStateChartTransition *)transition {
    NSString *eventKey = NSStringFromSelector(transition.event);

    NSMutableArray *transitions = [_events objectForKey:eventKey] ?: [NSMutableArray array];
    [transitions addObject:transition];
    [_events setObject:transitions forKey:eventKey];
}

- (void)onTimeout:(NSNumber *)timeout transitionTo:(SHPState *)toState event:(SEL)event guard:(SHPStateChartGuard *)guard historyMode:(HistoryMode)historyMode {
    // An action currently needs to be specified and defined in the state charts event port protocol
    // Creating an interface with no action specified is not so easy, because we want to use the timeout trigger to invoke the event on the machine and we need
    // the event port for doing this
    SHPStateChartTimeoutTransition *transition = [SHPStateChartTimeoutTransition transitionWithEvent:event fromStates:@[ self ] toStates:@[ toState ] timeout:timeout internal:NO guard:guard historyMode:historyMode];
    [self addEventTransition:transition];
}

- (void)deferEvent:(SEL)event {
    SHPStateChartDeferredTransition *transition = [[SHPStateChartDeferredTransition alloc] initWithEvent:event fromStates:@[self] toStates:nil timeout:nil internal:NO guard:nil historyMode:HistoryModeNone];
    [self addEventTransition:transition];
}

#pragma mark - Internal transitions

- (void)internalTransitionOnEvent:(SEL)event guard:(SHPStateChartGuard *)guard timeout:(NSNumber *)timeout {
    SHPStateChartTransition *transition = [SHPStateChartTransition transitionWithEvent:event fromStates:@[ self ] toStates:@[ ] timeout:timeout internal:YES guard:guard historyMode:HistoryModeNone];
    [self addEventTransition:transition];
}

- (void)internalTransitionOnEvent:(SEL)event {
    [self internalTransitionOnEvent:event guard:nil timeout:nil];
}

#pragma mark - Transitions

- (NSOrderedSet *)transitionsRespondingToEventMessage:(SHPEventMessage *)eventMessage currentPerformingTransition:(SHPStateChartTransition *)currentPerformingTransition {
    NSAssert(false, @"Must implement in subclass");
    return nil;
}

- (BOOL)isOrContainsTargetStates:(NSArray *)states {
    NSAssert(false, @"Must implement in subclass");
    return NO;
}

- (void)exitCurrentConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    NSAssert(false, @"Must implement in subclass");
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    NSAssert(false, @"Must implement in subclass");
}

- (void)enterDefaultConfigurationWithEventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    NSAssert(false, @"Must implement in subclass");
}


- (void)enterTargetConfigurationWithTargetStates:(NSArray *)targetStates eventMessage:(SHPEventMessage *)eventMessage enterSelf:(BOOL)enterSelf historyMode:(HistoryMode)historyMode transitionRepresentation:(SHPTransitionRepresentation *)eventTransitionRepresentation {
    NSAssert(false, @"Must implement in subclass");
}

- (void)enterDeepHistoryWithEventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    NSAssert(false, @"Must implement in subclass");
}

- (NSArray *)transitionsForEvent:(SEL)event checkCondition:(BOOL)condition {
    NSArray *transitions = _events[NSStringFromSelector(event)];
    if (condition) {
        return [transitions select:^BOOL(SHPStateChartTransition *transition) {
            if (transition.guard) {
                return transition.guard.condition();
            } else {
                return YES;
            }
        }];
    } else {
        return transitions;
    }
}

- (BOOL)shouldDeferEvent:(SEL)event {
    return [_deferEvents objectForKey:NSStringFromSelector(event)] != nil;
}

#pragma mark - Timeouts

- (void)setupTimeoutTimers {
    NSArray *timeoutTransitions = [[[_events allValues] arrayByFlattening] select:^BOOL(SHPStateChartTransition *transition) { return [transition isKindOfClass:[SHPStateChartTimeoutTransition class]]; }];

    _timeoutTimers = [timeoutTransitions collect:^id(SHPStateChartTimeoutTransition *transition) {
        MSWeakTimer *timer = [MSWeakTimer scheduledTimerWithTimeInterval:[transition.timeout doubleValue] target:self selector:@selector(eventTimeout:) userInfo:transition repeats:NO dispatchQueue:self.stateChart.messageQueue];
        return timer;
    }];
}

- (void)removeTimeoutTimers {
    for (NSTimer *timer in _timeoutTimers) {
        [timer invalidate];
    }
    _timeoutTimers = nil;
}

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
- (void)eventTimeout:(NSTimer *)timer {
    SHPStateChartTimeoutTransition *transition = timer.userInfo;
    [self.stateChart.eventPort performSelector:transition.event];
}
#pragma clang diagnostic pop

#pragma mark - Events

- (void)eventEntry:(SHPEventMessage *)eventMessage {
}

- (void)eventExit:(SHPEventMessage *)eventMessage {
}

- (void)eventInitial:(SHPEventMessage *)eventMessage {
}

- (void)guardTransition {
    // Just here to avoid warning in other method using it
}

#pragma mark - Representation

- (SHPStateRepresentation *)representationForState {
    NSAssert(false, @"Must implement in subclass");
    return nil;
}

@end
