// This software is licensed under GNU GLP version 3. Please read the file licence.txt for the complete licence text.
// 
// Copyright (c) 2014, Peter Gammelgaard, Philip Bruce & Shape A/S. All rights reserved.


#import <SHPStateChart/SHPEventRepresentation.h>
#import "SHPOrState.h"
#import "SHPEventRepresentation.h"
#import "SHPStateChart.h"
#import "SHPEventMessage.h"
#import "SHPStateChartTransition.h"
#import "SHPEventHandlingResult.h"
#import "SHPStateChartGuardTransition.h"
#import "SHPStateChartDeferredTransition.h"
#import "SHPOrStateRepresentation.h"
#import "SHPStateRepresentation.h"
#import "SHPStateChartRepresentation.h"
#import "SHPTransitionRepresentation.h"

@implementation SHPStateChartConfiguration

- (id)initWithRootState:(SHPOrState *)rootState {
    if (!(self = [super init])) { return nil; }

    _rootState = rootState;

    return self;
}
@end

@implementation SHPStateChart {
    NSMutableArray *_deferredEvents;
}

#pragma mark - Init

- (id)init {
    return [self initWithRootState:[SHPOrState new] skipRoot:NO];
}

- (id)initWithRootState:(SHPOrState *)rootState {
    return [self initWithRootState:rootState skipRoot:YES];
}

- (id)initWithRootState:(SHPOrState *)rootState skipRoot:(BOOL)skipRoot {
    self = [super init];
    if (self) {
        _rootState = rootState;

        SHPStateChartConfiguration *configuration = [[SHPStateChartConfiguration alloc] initWithRootState:_rootState];
        [self configureWithConfiguration:configuration];

        if (skipRoot) {
            [_rootState.subStates enumerateObjectsUsingBlock:^(SHPState *state, NSUInteger idx, BOOL *stop) {
                [state setStateChartUnlessSetRecursively:self];
            }];
        } else {
            // Set link to statechart on all states
            [_rootState setStateChartUnlessSetRecursively:self];
        }

        _eventPort = [[SHPEventPort alloc] initWithComponent:self protocol:configuration.eventHandlerProtocol];
        _stateChartInfoPort = [self outputPortWithProtocol:@protocol(SHPStateChartInfoProtocol) name:@"StateRecording"];
        _deferredEvents = [NSMutableArray array];
    }

    return self;
}

#pragma mark - Engine

/**
 *  The initial transition
 *
 *  This method must be called before any other in the state machine.
 *  It executes the state machines initial transition,
 *  and enters the correct leaf state after performing the transition
 */
- (void)initialTransition {
    dispatch_async(self.messageQueue, ^{
        // ----------------
        // Recursively enter initial states
        // ----------------
        SHPEventHandlingResult *eventHandlingResult = [[SHPEventHandlingResult alloc] initWithEvent:nil];
        SHPTransitionRepresentation *transitionRepresentation = [[SHPTransitionRepresentation alloc] initWithEvent:nil fromStates:nil toStates:nil timeout:@(0) internal:NO guard:nil historyMode:nil];
        [eventHandlingResult addTransitionRepresentation:transitionRepresentation];

        [_rootState enterDefaultConfigurationWithEventMessage:nil enterSelf:NO transitionRepresentation:transitionRepresentation];
        [self.stateChartInfoPort didAddInitialEventHandlingResult:eventHandlingResult];

        // ----------------
        // Check for transient / guard transitions
        // ----------------
        SHPEventMessage *eventMessage = [[SHPEventMessage alloc] initWithEventType:@selector(guardTransition) invocation:nil];
        [self performEvent:eventMessage];
    });
}

/**
 *  Find the Lowest Common Ancestor (LCA) between two states
 *
 *  Walks the tree backwards from source state to top state and from target state to top state.
 *  Then intersects the trees to find the LCA
 */
- (SHPOrState *)LCAFromSourceState:(SHPState *)sourceState targetState:(SHPState *)targetState {
    // ---------------------
    // Tree from source state to top state
    // ---------------------
    NSMutableOrderedSet *sourceSet = [[self statesUntilTopStateFromState:sourceState] mutableCopy];

    // ---------------------
    // Tree from target state to top state
    // ---------------------
    NSMutableOrderedSet *targetSet = [[self statesUntilTopStateFromState:targetState] mutableCopy];

    // ---------------------
    // Determine LCA
    // ---------------------
    NSMutableOrderedSet *LCASet = [targetSet mutableCopy];
    [LCASet intersectOrderedSet:sourceSet];
    SHPOrState *LCAState = [LCASet objectAtIndex:0];

    return LCAState;
}

/**
 *  Returns a set of states from the source state to the top state by walking up the state tree
*/
- (NSOrderedSet *)statesUntilTopStateFromState:(SHPState *)state {
    // ---------------------
    // Make state set by walking up the tree
    // ---------------------
    NSMutableOrderedSet *stateSet = [NSMutableOrderedSet orderedSet];
    [stateSet addObject:state];

    SHPState *superState = state;
    do {
        superState = superState.superstate;
        if(superState) {
            [stateSet addObject:superState];
        }
    } while (superState!=nil);

    return [stateSet copy];
}

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
/**
 *  Dispatches an event to the state machine
 *
 *  Performs all the steps necessary for delivering an event to the state machine:
 *  1. Find state which can handle the event
 *  2. Perform state actions
 *  3. If state results in a transition
 *  a. Exit source state to LCA
 *  b. Enter target state from LCA
 *  (exception) If the state handling the event equals the target state it is a self transition and the state is exited and entered even though it equals the LCA
 *  (exception) If the state target is also the state source but the state handling the state is further up the hierarchy the source state is exited until the performing state end then entered again
 *  c. Drill down state hierarchy if target state is a composite state
 *  d. Enter new target state
 *  e. Repeat (c) and (d) as necessary
 */
- (BOOL)performEvent:(SHPEventMessage *)eventMessage {
    @autoreleasepool {

        SHPEventHandlingResult *eventHandlingResult = [[SHPEventHandlingResult alloc] initWithEvent:NSStringFromSelector(eventMessage.eventType)];

        BOOL shouldCheckForGuardTransitions;
        do {
            shouldCheckForGuardTransitions = NO;
            // ---------------------
            // Perform state actions - walks up the tree until finding state that handles event
            // ---------------------
            NSOrderedSet *transitionsRespondingToEventMessage = [_rootState transitionsRespondingToEventMessage:eventMessage currentPerformingTransition:nil];

            for (SHPStateChartTransition *transition in transitionsRespondingToEventMessage) {
                // Representation
                SHPTransitionRepresentation *transitionRepresentation = [transition representationForTransition];
                [eventHandlingResult addTransitionRepresentation:transitionRepresentation];

                // Ensure guard transition check is run after handling these transitions
                shouldCheckForGuardTransitions = YES; // Should happen if at least one transition occurs

                if ([transition isInternal]) {
                    // Action
                    for (SHPState *state in transition.fromStates) {
                        [self actionForTransition:transition eventMessage:eventMessage state:state transitionRepresentation:transitionRepresentation];
                    }
                } else if ([transition isKindOfClass:[SHPStateChartDeferredTransition class]]) {
                    [_deferredEvents addObject:eventMessage];
                } else {
                    // Check for self transition
                    if (([transition.fromStates count] == 1 && [transition.toStates count] == 1) && (transition.fromStates[0] == transition.toStates[0])) {
                        SHPState *state = (SHPState *)transition.fromStates[0];
                        // Exit, Action, Enter
                        [state exitCurrentConfigurationWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
                        [self actionForTransition:transition eventMessage:eventMessage state:state transitionRepresentation:transitionRepresentation];
                        [state enterTargetConfigurationWithTargetStates:@[ state ] eventMessage:eventMessage enterSelf:YES historyMode:HistoryModeNone transitionRepresentation:transitionRepresentation];
                    } else {
                        SHPOrState *LCA = [self LCAFromSourceState:transition.fromStates[0] targetState:transition.toStates[0]];
                        // Exit, Action, Enter
                        [LCA.activeState exitCurrentConfigurationWithEventMessage:eventMessage transitionRepresentation:transitionRepresentation];
                        for (SHPState *state in transition.fromStates) {
                            [self actionForTransition:transition eventMessage:eventMessage state:state transitionRepresentation:transitionRepresentation];
                        }
                        [self enterTargetConfigurationWithState:LCA transition:transition eventMessage:eventMessage transitionRepresentation:transitionRepresentation];

                        // Replay any deferred events on the event queue since we have now changed state
                        for (SHPEventMessage *msg in _deferredEvents) {
                            dispatch_async(self.messageQueue, ^{
                                [self performEvent:msg];
                            });
                        }
                        [_deferredEvents removeAllObjects];
                    }
                }
            }

            if (shouldCheckForGuardTransitions) {
                eventMessage = [[SHPEventMessage alloc] initWithEventType:@selector(guardTransition) invocation:nil];
            }

        } while (shouldCheckForGuardTransitions);

        // ----------------
        // Send event changes
        // ----------------
        [self.stateChartInfoPort didAddEventHandlingResult:eventHandlingResult];

        // ----------------
        // Send updated state
        // ----------------
        SHPOrStateRepresentation *representationRootState = [_rootState representationForState];
        SHPStateChartRepresentation *stateChartRepresentation = [[SHPStateChartRepresentation alloc] init];
        stateChartRepresentation.rootState = representationRootState;
        [self.stateChartInfoPort didUpdateStateChartRepresentation:stateChartRepresentation];

        return YES;
    }

}

- (void)enterTargetConfigurationWithState:(SHPOrState *)state transition:(SHPStateChartTransition *)transition eventMessage:(SHPEventMessage *)eventMessage transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    [state enterTargetConfigurationWithTargetStates:transition.toStates eventMessage:eventMessage enterSelf:NO historyMode:transition.historyMode transitionRepresentation:transitionRepresentation];
}

- (void)actionForTransition:(SHPStateChartTransition *)transition eventMessage:(SHPEventMessage *)eventMessage state:(SHPState *)state transitionRepresentation:(SHPTransitionRepresentation *)transitionRepresentation {
    SHPEventRepresentation *eventRepresentation = [[SHPEventRepresentation alloc] initWithEventType:(SHPEventRepresentationEventTypeAction) transitionAction:NSStringFromSelector(eventMessage.eventType) stateRepresentation:[state representationForState]];
    [transitionRepresentation addEventRepresentation:eventRepresentation];

    if ([transition isKindOfClass:[SHPStateChartGuardTransition class]]) {
        SEL action = [(SHPStateChartGuardTransition *)transition action];
        if ([state respondsToSelector:action]) {
            [state performSelector:action];
        }
    } else {
        if([state respondsToSelector:eventMessage.eventType]) {
            [self invokeEventMethodFromEventMessage:eventMessage state:state];
        }
    }
}

- (void)invokeEventMethodFromEventMessage:(SHPEventMessage *)eventMessage state:(SHPState *)state {
    NSInvocation *invocation = eventMessage.invocation;
    [invocation setTarget:state];
    [invocation invoke];
}

#pragma clang diagnostic pop

#pragma mark - Port delegate

- (void)inputPort:(SHPInputPort *)inputPort didReceiveMessage:(SHPEventMessage *)message {
    [self performEvent:message];
}

#pragma mark -

- (SHPState *)findStateOfClass:(Class)stateClass {
    return [_rootState findFirstStateOfClassRecursively:(Class)stateClass];
}

@end
